{% extends "!layout.html" %}

{% block htmltitle %}
<!-- Google tag (gtag.js) -->
<script async src="https://www.googletagmanager.com/gtag/js?id=G-G8M9T0EGN0"></script>
<script>
    window.dataLayer = window.dataLayer || [];
    function gtag() { dataLayer.push(arguments); }
    gtag('js', new Date());

    gtag('config', 'G-G8M9T0EGN0');
</script>
{% if pagename == 'index' %}
<title>DataLab</title>
{% else %}
{{ super() }}
{% endif %}
{% endblock %}